# -*- coding: utf-8 -*-
from typing import List, Optional

import attr
import os

from .const import REGEX_FASTQ


@attr.s
class Subtype(object):
    sample = attr.ib(validator=attr.validators.instance_of(str))
    file_path = attr.ib()
    scheme = attr.ib(validator=attr.validators.instance_of(str))
    scheme_version = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    subtype = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    non_present_subtypes = attr.ib(default=None)  # type: Optional[List[str]]
    all_subtypes = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    inconsistent_subtypes = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    tiles_matching_subtype = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    negative_tiles_matching_subtype = attr.ib(default=None,
                                              validator=attr.validators.optional(attr.validators.instance_of(str)))
    are_subtypes_consistent = attr.ib(default=True, validator=attr.validators.instance_of(bool))
    n_tiles_matching_all = attr.ib(default=0)
    n_tiles_matching_positive = attr.ib(default=0)
    n_tiles_matching_negative = attr.ib(default=0)
    n_tiles_matching_subtype = attr.ib(default=0)
    n_tiles_matching_all_expected = attr.ib(default=0)
    n_tiles_matching_positive_expected = attr.ib(default=0)
    n_tiles_matching_negative_expected = attr.ib(default=0)
    n_tiles_matching_subtype_expected = attr.ib(default=0)
    n_negative_tiles_matching_subtype_expected = attr.ib(default=0)
    avg_tile_coverage = attr.ib(default=None)
    qc_status = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    qc_message = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    scheme_subtype_counts = attr.ib(default=None, repr=False)

    @file_path.validator
    def _file_path_validator(self, attribute, value):
        if isinstance(value, str):
            if not os.path.exists(self.file_path):
                raise OSError('Input file "{}" does not exist!'.format(value))
        elif isinstance(value, list):
            for x in value:
                if not os.path.exists(x):
                    raise OSError('Input file "{}" does not exist!'.format(x))
        else:
            raise ValueError('Unexpected type for input file path "{}": {}'.format(type(value), value))

    def is_fastq_input(self):
        if isinstance(self.file_path, str):
            return bool(REGEX_FASTQ.match(self.file_path))
        elif isinstance(self.file_path, list):
            return all([bool(REGEX_FASTQ.match(x)) for x in self.file_path])
        else:
            raise ValueError('Unexpected type "{}" for "file_path": {}'.format(type(self.file_path), self.file_path))



# -*- coding: utf-8 -*-

class QC:
    FAIL = 'FAIL'
    WARNING = 'WARNING'
    PASS = 'PASS'
    # Errors for Hansel
    MISSING_TILES_ERROR_1 = 'Missing Tiles Error 1'
    MIXED_SAMPLE_ERROR_2 = 'Mixed Sample Error 2'
    AMBIGUOUS_RESULTS_ERROR_3 = 'Ambiguous Results Error 3'
    UNCONFIDENT_RESULTS_ERROR_4 = 'Unconfident Results Error 4'
    INTERMEDIATE_SUBTYPE_WARNING = 'Intermediate Subtype'
    LOW_COVERAGE_WARNING = 'Low Coverage'
    NO_SUBTYPE_RESULT = 'No subtype result!'



# -*- coding: utf-8 -*-

import logging

import re

VALID_NUCLEOTIDES = {'A', 'a',
                     'C', 'c',
                     'G', 'g',
                     'T', 't',
                     'R', 'r',
                     'Y', 'y',
                     'S', 's',
                     'W', 'w',
                     'K', 'k',
                     'M', 'm',
                     'B', 'b',
                     'D', 'd',
                     'H', 'h',
                     'V', 'v',
                     'N', 'n',
                     'X', 'x', }  # X for masked nucleotides

REGEX_GZIPPED = re.compile(r'^.+\.gz$')


def parse_fasta(filepath):
    '''Parse a FASTA/FASTA.GZ file returning a generator yielding tuples of fasta headers to sequences.

    Args:
        filepath (str): Fasta file path

    Returns:
        generator: yields tuples of (<fasta header>, <fasta sequence>)
    '''
    if REGEX_GZIPPED.match(filepath):
        logging.debug('Opening "%s" as gzipped file', filepath)
        # using os.popen with zcat since it is much faster than gzip.open or gzip.open(io.BufferedReader)
        # http://aripollak.com/pythongzipbenchmarks/
        # assumes Linux os with zcat installed
        import os
        with os.popen('zcat {}'.format(filepath)) as f:
            yield from _parse_fasta(f, filepath)
    else:
        with open(filepath, 'r') as f:
            yield from _parse_fasta(f, filepath)


def _parse_fasta(f, filepath):
    seqs = []
    header = ''
    line_count = 0
    for line in f:
        if isinstance(line, bytes):
            line = line.decode()
        line = line.strip()
        if line == '':
            continue
        if line[0] == '>':
            if header == '':
                header = line.replace('>', '')
            else:
                yield header, ''.join(seqs)
                seqs = []
                header = line.replace('>', '')
        else:
            non_nucleotide_chars_in_line = set(line) - VALID_NUCLEOTIDES
            if len(non_nucleotide_chars_in_line) > 0:
                msg = '{file}: Line {line} contains the following non-nucleotide characters: {chars}'.format(
                    file=filepath,
                    line=line_count,
                    chars=', '.join([str(x) for x in non_nucleotide_chars_in_line]))
                logging.warning(msg)
            seqs.append(line)
        line_count += 1
    yield header, ''.join(seqs)


def parse_fastq(filepath):
    """Parse a FASTQ/FASTQ.GZ file returning a generator yielding tuples of FASTQ entry headers and sequences.

    Args:
        filepath (str): FASTQ/FASTQ.GZ file path

    Returns:
        generator: yields tuples of (<fastq header>, <fastq sequence>)
    """
    if REGEX_GZIPPED.match(filepath):
        logging.debug('Opening "%s" as gzipped file', filepath)
        # using os.popen with zcat since it is much faster than gzip.open or gzip.open(io.BufferedReader)
        # http://aripollak.com/pythongzipbenchmarks/
        # assumes Linux os with zcat installed
        import os
        with os.popen('zcat {}'.format(filepath)) as f:
            yield from _parse_fastq(f)
    else:
        with open(filepath, 'rU') as f:
            yield from _parse_fastq(f)


def _parse_fastq(f):
    """Simple FASTQ parser which yields the header and sequence ignoring the quality scores

    Args:
        f: file-like object

    Yields:
        Tuple of FASTQ entry header and sequence
    """
    header = ''
    seq = ''
    skip = False
    for line in f:
        if skip:
            skip = False
            continue
        line = line.strip()
        if line == '':
            continue
        if line[0] == '@':
            header = line.replace('@', '')
        elif line[0] == '+':
            yield header, seq
            skip = True
        else:
            seq = line



# -*- coding: utf-8 -*-
import attr

@attr.s
class SubtypingParams(object):
    low_coverage_depth_freq = attr.ib(default=20.0, validator=attr.validators.instance_of((float, int)))
    max_perc_missing_tiles = attr.ib(default=0.05, validator=attr.validators.instance_of(float))
    min_ambiguous_tiles = attr.ib(default=3, validator=attr.validators.instance_of(int))
    max_perc_intermediate_tiles = attr.ib(default=0.05, validator=attr.validators.instance_of(float))
    min_kmer_freq = attr.ib(default=8, validator=attr.validators.instance_of((float, int)))
    max_kmer_freq = attr.ib(default=1000, validator=attr.validators.instance_of((float, int)))
    min_coverage_warning = attr.ib(default=20, validator=attr.validators.instance_of((float, int)))

    @max_perc_missing_tiles.validator
    def _validate_max_perc_missing_tiles(self, attribute, value):
        if 0.0 > value > 1.0:
            raise AttributeError('Max % misssing tiles was {} expected to be decimal between 0.0 and 1.0 inclusive'.format(value))



# -*- coding: utf-8 -*-

import re
import logging
from typing import Optional, List, Dict, Union, Tuple

import attr
import pandas as pd

from .aho_corasick import init_automaton, find_in_fasta, find_in_fastqs
from .const import COLUMNS_TO_REMOVE
from .qc import perform_quality_check
from .subtype import Subtype
from .subtype_stats import SubtypeCounts
from .subtype_stats import subtype_counts
from .subtyping_params import SubtypingParams
from .utils import find_inconsistent_subtypes, get_scheme_fasta, get_scheme_version, init_subtyping_params


def subtype_contigs_ac(fasta_path: str,
                       genome_name: str,
                       scheme: str,
                       subtyping_params: Optional[SubtypingParams] = None,
                       scheme_name: Optional[str] = None,
                       scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None) -> Tuple[
    Subtype, Optional[pd.DataFrame]]:
    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    if subtyping_params is None:
        subtyping_params = init_subtyping_params(scheme=scheme)
    scheme_version = get_scheme_version(scheme)
    st = Subtype(sample=genome_name,
                 file_path=fasta_path,
                 scheme=scheme_name or scheme,
                 scheme_version=scheme_version,
                 scheme_subtype_counts=scheme_subtype_counts)

    A = init_automaton(scheme_fasta)
    df = find_in_fasta(A, fasta_path)

    if df is None or df.shape[0] == 0:
        logging.warning('No subtyping tile matches for input "%s" for scheme "%s"', fasta_path, scheme)
        st.are_subtypes_consistent = False
        return st, None

    refpositions = [x for x, y in df.tilename.str.split('-')]
    subtypes = [y for x, y in df.tilename.str.split('-')]
    df['refposition'] = [int(x.replace('negative', '')) for x in refpositions]
    df['subtype'] = subtypes
    df['is_pos_tile'] = ~df.tilename.str.contains('negative')
    dfpos = df[df.is_pos_tile]
    subtype_lens = dfpos.subtype.apply(len)
    max_subtype_strlen = subtype_lens.max()
    logging.debug('max substype str len: %s', max_subtype_strlen)
    dfpos_highest_res = dfpos[subtype_lens == max_subtype_strlen]
    logging.debug('dfpos_highest_res: %s', dfpos_highest_res)
    pos_subtypes = [[int(y) for y in x.split('.')] for x in dfpos.subtype.unique()]
    pos_subtypes.sort(key=lambda a: len(a))
    logging.debug('pos_subtypes: %s', pos_subtypes)
    inconsistent_subtypes = find_inconsistent_subtypes(pos_subtypes)
    logging.debug('inconsistent_subtypes: %s', inconsistent_subtypes)
    st.n_tiles_matching_all = df.tilename.unique().size
    st.n_tiles_matching_positive = dfpos.tilename.unique().size
    st.n_tiles_matching_subtype = dfpos_highest_res.tilename.unique().size
    pos_subtypes_str = [x for x in dfpos.subtype.unique()]
    pos_subtypes_str.sort(key=lambda x: len(x))
    st.all_subtypes = '; '.join(pos_subtypes_str)
    subtype_list = [x for x in dfpos_highest_res.subtype.unique()]
    st.subtype = '; '.join(subtype_list)
    st.n_tiles_matching_all_expected = ';'.join([str(scheme_subtype_counts[x].all_tile_count) for x in subtype_list])
    st.n_tiles_matching_positive_expected = ';'.join(
        [str(scheme_subtype_counts[x].positive_tile_count) for x in subtype_list])
    st.n_tiles_matching_subtype_expected = ';'.join(
        [str(scheme_subtype_counts[x].subtype_tile_count) for x in subtype_list])
    st.tiles_matching_subtype = '; '.join([x for x in dfpos_highest_res.tilename.unique()])

    if len(inconsistent_subtypes) > 0:
        st.are_subtypes_consistent = False
        st.inconsistent_subtypes = inconsistent_subtypes

    possible_downstream_subtypes = [s for s in scheme_subtype_counts
                                    if re.search("^({})(\.)(\d)$".format(re.escape(st.subtype)), s)]
    st.non_present_subtypes = [x for x in possible_downstream_subtypes
                               if not (df.subtype == x).any()]
    st.qc_status, st.qc_message = perform_quality_check(st, df, subtyping_params)

    logging.info(st)

    df['sample'] = genome_name
    df['file_path'] = fasta_path
    df['scheme'] = scheme_name or scheme
    df['scheme_version'] = scheme_version
    df['qc_status'] = st.qc_status
    df['qc_message'] = st.qc_message

    df = df[df.columns[~df.columns.isin(COLUMNS_TO_REMOVE)]]
    return st, df


def parallel_query_contigs_ac(input_genomes: List[Tuple[str, str]],
                              scheme: str,
                              scheme_name: Optional[str] = None,
                              subtyping_params: Optional[SubtypingParams] = None,
                              scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                              n_threads: int = 1):
    from multiprocessing import Pool
    logging.info('Initializing thread pool with %s threads', n_threads)
    pool = Pool(processes=n_threads)
    logging.info('Running analysis asynchronously on %s input genomes', len(input_genomes))
    res = [pool.apply_async(subtype_contigs_ac, (input_fasta,
                                                 genome_name,
                                                 scheme,
                                                 subtyping_params,
                                                 scheme_name,
                                                 scheme_subtype_counts))
           for input_fasta, genome_name in input_genomes]
    logging.info('Parallel analysis complete! Retrieving analysis results')
    outputs = [x.get() for x in res]
    return outputs


def query_contigs_ac(subtype_results: List[Subtype],
                     dfs: List[pd.DataFrame],
                     input_genomes: List[Tuple[str, str]],
                     scheme: str,
                     scheme_name: Optional[str] = None,
                     subtyping_params: Optional[SubtypingParams] = None,
                     scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                     n_threads: int = 1) -> None:
    if n_threads == 1:
        logging.info('Serial single threaded run mode on %s input genomes', len(input_genomes))
        outputs = [subtype_contigs_ac(fasta_path=input_fasta,
                                      genome_name=genome_name,
                                      scheme=scheme,
                                      scheme_name=scheme_name,
                                      subtyping_params=subtyping_params,
                                      scheme_subtype_counts=scheme_subtype_counts)
                   for input_fasta, genome_name in input_genomes]
    else:
        outputs = parallel_query_contigs_ac(input_genomes, scheme, scheme_name, subtyping_params, scheme_subtype_counts,
                                            n_threads)
    for subtype, df in outputs:
        if df is not None:
            dfs.append(df)
        subtype_results.append(attr.asdict(subtype))


def parallel_query_reads_ac(reads: List[Tuple[List[str], str]],
                            scheme: str,
                            scheme_name: Optional[str] = None,
                            subtyping_params: Optional[SubtypingParams] = None,
                            scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                            n_threads: int = 1) -> List[Tuple[Subtype, pd.DataFrame]]:
    from multiprocessing import Pool
    logging.info('Initializing thread pool with %s threads', n_threads)
    pool = Pool(processes=n_threads)
    logging.info('Running analysis asynchronously on %s input genomes', len(reads))
    res = [pool.apply_async(subtype_reads_ac, (fastqs,
                                               genome_name,
                                               scheme,
                                               scheme_name,
                                               subtyping_params,
                                               scheme_subtype_counts))
           for fastqs, genome_name in reads]
    logging.info('Parallel analysis complete! Retrieving analysis results')
    outputs = [x.get() for x in res]
    return outputs


def subtype_reads_ac(reads: Union[str, List[str]],
                     genome_name: str,
                     scheme: str,
                     scheme_name: Optional[str] = None,
                     subtyping_params: Optional[SubtypingParams] = None,
                     scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None) -> Tuple[
    Subtype, Optional[pd.DataFrame]]:
    logging.info("genome_name %s", genome_name)

    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    if subtyping_params is None:
        subtyping_params = init_subtyping_params(scheme=scheme)
    scheme_version = get_scheme_version(scheme)

    st = Subtype(sample=genome_name,
                 file_path=reads,
                 scheme=scheme_name or scheme,
                 scheme_version=scheme_version,
                 scheme_subtype_counts=scheme_subtype_counts)

    A = init_automaton(scheme_fasta)
    if isinstance(reads, str):
        df = find_in_fastqs(A, reads)
    elif isinstance(reads, list):
        df = find_in_fastqs(A, *reads)
    else:
        raise ValueError('Unexpected type "{}" for "reads": {}'.format(type(reads), reads))

    if df is None or df.shape[0] == 0:
        logging.warning('No subtyping tile matches for input "%s" for scheme "%s"', reads, scheme)
        st.are_subtypes_consistent = False
        return st, None

    refpositions = [x for x, y in df.tilename.str.split('-')]
    subtypes = [y for x, y in df.tilename.str.split('-')]
    df['refposition'] = [int(x.replace('negative', '')) for x in refpositions]
    df['subtype'] = subtypes
    df['is_pos_tile'] = ~df.tilename.str.contains('negative')
    df['is_kmer_freq_okay'] = (df.freq >= subtyping_params.min_kmer_freq) & (df.freq <= subtyping_params.max_kmer_freq)
    dfgood = df[df.is_kmer_freq_okay]
    dfpos = dfgood[dfgood.is_pos_tile]
    dfneg = dfgood[~dfgood.is_pos_tile]
    subtype_lens = dfpos.subtype.apply(len)
    max_subtype_strlen = subtype_lens.max()
    logging.debug('max substype str len: %s', max_subtype_strlen)
    dfpos_highest_res = dfpos[subtype_lens == max_subtype_strlen]
    pos_subtypes = [[int(y) for y in x.split('.')] for x in dfpos.subtype.unique()]
    pos_subtypes.sort(key=lambda a: len(a))
    logging.debug('pos_subtypes: %s', pos_subtypes)
    inconsistent_subtypes = find_inconsistent_subtypes(pos_subtypes)
    logging.debug('inconsistent_subtypes: %s', inconsistent_subtypes)
    st.n_tiles_matching_all = dfgood.shape[0]
    st.n_tiles_matching_positive = dfpos.shape[0]
    st.n_tiles_matching_negative = dfneg.shape[0]
    st.n_tiles_matching_subtype = dfpos_highest_res.shape[0]
    pos_subtypes_str = [x for x in dfpos.subtype.unique()]
    pos_subtypes_str.sort(key=lambda x: len(x))
    st.all_subtypes = '; '.join(pos_subtypes_str)
    subtype_list = [x for x in dfpos_highest_res.subtype.unique()]
    st.subtype = '; '.join(subtype_list)
    st.n_tiles_matching_all_expected = ';'.join([str(scheme_subtype_counts[x].all_tile_count) for x in subtype_list])
    st.n_tiles_matching_positive_expected = ';'.join(
        [str(scheme_subtype_counts[x].positive_tile_count) for x in subtype_list])
    st.n_tiles_matching_subtype_expected = ';'.join(
        [str(scheme_subtype_counts[x].subtype_tile_count) for x in subtype_list])
    st.tiles_matching_subtype = '; '.join([x for x in dfpos_highest_res.tilename.unique()])
    if len(inconsistent_subtypes) > 0:
        st.are_subtypes_consistent = False
        st.inconsistent_subtypes = inconsistent_subtypes
    possible_downstream_subtypes = [s for s in scheme_subtype_counts
                                    if re.search("^({})(\.)(\d)$".format(re.escape(st.subtype)), s)]
    st.non_present_subtypes = [x for x in possible_downstream_subtypes
                               if not (df.subtype == x).any()]
    st.avg_tile_coverage = df['freq'].mean()
    st.qc_status, st.qc_message = perform_quality_check(st, df, subtyping_params)
    df['sample'] = genome_name
    df['scheme'] = scheme_name or scheme
    df['scheme_version'] = scheme_version
    df['qc_status'] = st.qc_status
    df['qc_message'] = st.qc_message
    df = df[df.columns[~df.columns.isin(COLUMNS_TO_REMOVE)]]
    return st, df


def query_reads_ac(subtype_results: List[Subtype],
                   dfs: List[pd.DataFrame],
                   reads: List[Tuple[List[str], str]],
                   scheme: str,
                   scheme_name: Optional[str] = None,
                   subtyping_params: Optional[SubtypingParams] = None,
                   scheme_subtype_counts: Optional[Dict[str, SubtypeCounts]] = None,
                   n_threads: int = 1):
    if n_threads == 1:
        logging.info('Serial single threaded run mode on %s input genomes', len(reads))
        outputs = [subtype_reads_ac(reads=fastq_files,
                                    genome_name=genome_name,
                                    scheme=scheme,
                                    scheme_name=scheme_name,
                                    subtyping_params=subtyping_params,
                                    scheme_subtype_counts=scheme_subtype_counts)
                   for fastq_files, genome_name in reads]
    else:
        outputs = parallel_query_reads_ac(reads=reads,
                                          scheme=scheme,
                                          scheme_name=scheme_name,
                                          subtyping_params=subtyping_params,
                                          scheme_subtype_counts=scheme_subtype_counts,
                                          n_threads=n_threads)
    for subtype, df in outputs:
        if df is not None:
            dfs.append(df)
        subtype_results.append(attr.asdict(subtype))



# -*- coding: utf-8 -*-

from typing import Tuple, Optional

from pandas import DataFrame

from ..subtyping_params import SubtypingParams
from ..qc.const import QC
from ..qc.utils import get_conflicting_tiles, get_num_pos_neg_tiles, get_mixed_subtype_tile_counts
from ..subtype import Subtype


def is_overall_coverage_low(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    if not st.are_subtypes_consistent \
            or st.subtype is None \
            or not st.is_fastq_input():
        return None, None

    if st.avg_tile_coverage < p.min_coverage_warning:
        return QC.WARNING, '{}: Low coverage for all tiles ({:.3f} < {} expected)'.format(
            QC.LOW_COVERAGE_WARNING,
            st.avg_tile_coverage,
            p.min_coverage_warning
        )

    return None, None


def is_missing_tiles(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """Are there more missing tiles than tolerated?

    Note:
        For reads, calculate the average coverage depth from the tiles that are present and provide an adequate
        error message based on the coverage.

    Args:
        st: Subtype results
        df: Subtyping results dataframe
        p: Subtyping/QC parameters specifically `max_perc_missing_tiles` for % missing tiles threshold

    Returns:
        None, None if less missing tiles than tolerate; otherwise, "FAIL", error message
    """

    if st.are_subtypes_consistent:
        return check_for_missing_tiles(is_fastq=st.is_fastq_input(),
                                       curr_subtype=st.subtype,
                                       scheme=st.scheme,
                                       df=df,
                                       exp=int(st.n_tiles_matching_all_expected),
                                       obs=int(st.n_tiles_matching_all),
                                       p=p)
    else:
        message_list = []

        subtype_list = st.subtype.split(';')
        n_tiles_matching_expected = st.n_tiles_matching_all_expected.split(';')

        dfpos = df[df.is_pos_tile]
        mixed_subtype_counts = get_mixed_subtype_tile_counts(dfpos=dfpos,
                                                             subtype_list=subtype_list)

        for curr_subtype, exp in zip(subtype_list, n_tiles_matching_expected):
            # We can omit the status because there will be a fail status already from non consistent subtypes.
            _, curr_messages = check_for_missing_tiles(is_fastq=st.is_fastq_input(),
                                                       curr_subtype=curr_subtype,
                                                       scheme=st.scheme,
                                                       df=df,
                                                       exp=int(exp),
                                                       obs=(mixed_subtype_counts.get(curr_subtype) +
                                                            int(st.n_tiles_matching_negative)),
                                                       p=p)

            message_list.append(curr_messages)

        error_messages = ' | '.join(filter(None.__ne__, message_list))

        return _, error_messages


def check_for_missing_tiles(is_fastq: bool, curr_subtype: str, scheme: str,
                            df: DataFrame, exp: int, obs: int, p: SubtypingParams):
    error_status = None
    error_messages = None

    p_missing = (exp - obs) / exp  # type: float
    if p_missing > p.max_perc_missing_tiles:
        if is_fastq:
            tiles_with_hits = df[df['is_kmer_freq_okay']]  # type: DataFrame
            depth = tiles_with_hits['freq'].mean()
            if depth < p.low_coverage_depth_freq:
                coverage_msg = 'Low coverage depth ({:.1f} < {:.1f} expected); you may need more WGS data.'.format(
                    depth,
                    float(p.low_coverage_depth_freq))
            else:
                coverage_msg = 'Okay coverage depth ({:.1f} >= {:.1f} expected), but this may be the wrong ' \
                               'serovar or species for scheme "{}"'.format(
                                depth,
                                float(p.low_coverage_depth_freq),
                                scheme)
            error_messages = '{status}: {p_missing:.2%} missing tiles; more than {p_missing_threshold:.2%} missing ' \
                             'tiles threshold. {coverage_msg}'.format(
                                status=QC.MISSING_TILES_ERROR_1,
                                p_missing=p_missing,
                                p_missing_threshold=p.max_perc_missing_tiles,
                                coverage_msg=coverage_msg)
        else:
            error_messages = '{}: {:.2%} missing tiles for subtype {}; more than {:.2%} missing tile threshold'.format(
                QC.MISSING_TILES_ERROR_1,
                p_missing,
                curr_subtype,
                p.max_perc_missing_tiles)
        error_status = QC.FAIL

    return error_status, error_messages


def is_mixed_subtype(st: Subtype, df: DataFrame, *args) -> Tuple[Optional[str], Optional[str]]:
    """Is the subtype result mixed?

    Note:
        Check for `+` and `-` tiles that are present for the target site above minimum coverage threshold.
        If the subtyping result came back with an inconsistent call, provide an adequate error message to the user.

    Args:
        st: Subtype results
        df: Subtyping results dataframe
        args: unused args

    Returns:
        None, None if not mixed subtype result; otherwise, "FAIL", error message
    """
    if not st.are_subtypes_consistent:
        return QC.FAIL, \
               '{}: Mixed subtypes found: "{}".'.format(
                   QC.MIXED_SAMPLE_ERROR_2,
                   '; '.join(sorted(st.inconsistent_subtypes)))
    conflicting_tiles = get_conflicting_tiles(st, df)
    if conflicting_tiles is None or conflicting_tiles.shape[0] == 0:
        return None, None

    return QC.FAIL, '{}: Mixed subtype detected. Positive and negative tiles detected for ' \
                    'the same target site "{}" for subtype "{}".'.format(
                        QC.MIXED_SAMPLE_ERROR_2,
                        '; '.join(conflicting_tiles['refposition'].astype(str).tolist()),
                        st.subtype)


def is_missing_too_many_target_sites(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[
    Optional[str], Optional[str]]:
    """Are there too many missing target sites for an expected subtype?

    Check if there are any refpositions missing from the subtyping scheme in the result.
    If there are missing refpositions, this is a missing target.
    If there are too many missing targets, this result is an Ambiguous Result.

    Args:
        st: Subtype results
        df: Subtyping results dataframe
        p: Subtyping/QC parameters

    Returns:
        None, None if less missing targets than tolerated; otherwise, "FAIL", error message
    """
    if not st.are_subtypes_consistent \
            or st.subtype is None:
        return None, None

    potential_subtypes = st.all_subtypes.split('; ')
    uniq_positions = {y for x in potential_subtypes for y in st.scheme_subtype_counts[x].refpositions}
    missing_targets = [x for x in uniq_positions if (df.refposition == x).sum() == 0]

    exp = int(st.n_tiles_matching_all_expected)
    obs = int(st.n_tiles_matching_all)
    if (exp - obs) / exp <= p.max_perc_missing_tiles and len(missing_targets) >= p.min_ambiguous_tiles:
        return QC.FAIL, \
               '{}: {} missing positions detected for subtype: {}.'.format(
                   QC.AMBIGUOUS_RESULTS_ERROR_3,
                   len(missing_targets),
                   st.subtype)
    return None, None


def is_missing_downstream_targets(st: Subtype, *args) -> Tuple[Optional[str], Optional[str]]:
    """Are there missing downstream targets?

    Note:
        This method will check if there's any non_present_subtypes in the result, which would indicate a non confident
        result. This is due to the fact if you have a subtyping result of `2.1.1.2` and you're missing `2.1.1.2.X`
        You can't be sure that the subtype's final call is 2.1.1.2 if you're missing information.

    Args:
        st: Subtype results
        args: unused extra args

    Returns:
        None, None if no missing downstream targets; otherwise, "FAIL", error message
    """
    if st.non_present_subtypes:
        return QC.FAIL, \
               '{}: Subtype {} was found, but tiles for downstream subtype(s) "{}" were missing. ' \
               'Due to missing downstream tiles, there is a lack of confidence in the final subtype call.'.format(
                   QC.UNCONFIDENT_RESULTS_ERROR_4,
                   st.subtype,
                   st.non_present_subtypes)

    return None, None


def is_maybe_intermediate_subtype(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[
    Optional[str], Optional[str]]:
    '''Is the result a possible intermediate subtype?

    Return a WARNING message if all the conditions are true:
    - 95% of the scheme tiles are found
    - 0 conflicting tiles
    - total subtype tiles < expected
    - +/- tiles exist in the result.

    Args:
        st: Subtype results
        df: Subtyping results dataframe
        p: Subtyping/QC parameters specifically using `max_perc_intermediate_tiles`

    Returns:
        None, None if no intermediate subtype possible; otherwise, "FAIL", error message
    '''
    if not st.are_subtypes_consistent \
            or st.subtype is None:
        return None, None

    total_subtype_tiles = int(st.n_tiles_matching_subtype_expected)
    total_subtype_tiles_hits = int(st.n_tiles_matching_subtype)
    conflicting_tiles = get_conflicting_tiles(st, df)
    num_pos_tiles, num_neg_tiles = get_num_pos_neg_tiles(st, df)
    obs = int(st.n_tiles_matching_all)
    exp = int(st.n_tiles_matching_all_expected)
    if (exp - obs) / exp <= p.max_perc_intermediate_tiles and conflicting_tiles.shape[0] == 0 and \
                    total_subtype_tiles_hits < total_subtype_tiles and num_pos_tiles and num_neg_tiles:
        return QC.WARNING, \
               '{}: Possible intermediate subtype. All scheme tiles were found, but a fraction ' \
               'were positive for the final subtype. Total subtype matches observed (n={}) vs expected (n={})'.format(
                   QC.INTERMEDIATE_SUBTYPE_WARNING,
                   total_subtype_tiles_hits,
                   total_subtype_tiles)
    return None, None



# -*- coding: utf-8 -*-

from collections import defaultdict

from ahocorasick import Automaton
import pandas as pd

from ..parsers import parse_fasta, parse_fastq
from ..utils import revcomp


def init_automaton(scheme_fasta):
    """Initialize Aho-Corasick Automaton with kmers from SNV scheme fasta

    Args:
        scheme_fasta: SNV scheme fasta file path

    Returns:
         Aho-Corasick Automaton with kmers loaded
    """
    A = Automaton()
    for header, sequence in parse_fasta(scheme_fasta):
        A.add_word(sequence, (header, sequence, False))
        A.add_word(revcomp(sequence), (header, sequence, True))
    A.make_automaton()
    return A


def find_in_fasta(A: Automaton, fasta: str) -> pd.DataFrame:
    """Find scheme kmers in input fasta file

    Args:
        A: Aho-Corasick Automaton with scheme SNV target kmers loaded
        fasta: Input fasta path

    Returns:
        Dataframe with any matches found in input fasta file
    """
    res = []
    for contig_header, sequence in parse_fasta(fasta):
        for idx, (tilename, tile_seq, is_revcomp) in A.iter(sequence):
            res.append((tilename, tile_seq, is_revcomp, contig_header, idx))
    df = pd.DataFrame(res, columns=['tilename', 'seq', 'is_revcomp', 'contig_id', 'match_index'])
    return df


def find_in_fastqs(A: Automaton, *fastqs):
    """Find scheme kmers in input fastq files

    Args:
        A: Aho-Corasick Automaton with scheme SNV target kmers loaded
        fastqs: Input fastq file paths

    Returns:
        Dataframe with any matches found in input fastq files
    """
    tile_seq_counts = defaultdict(int)
    for fastq in fastqs:
        for _, sequence in parse_fastq(fastq):
            for idx, (_, tile_seq, _) in A.iter(sequence):
                tile_seq_counts[tile_seq] += 1
    res = []
    for tile_seq, freq in tile_seq_counts.items():
        tilename, sequence, _ = A.get(tile_seq)
        res.append((tilename, tile_seq, freq))
    df = pd.DataFrame(res, columns=['tilename', 'seq', 'freq'])
    return df



# -*- coding: utf-8 -*-

from typing import List, Callable, Tuple

from pandas import DataFrame

from ..subtype import Subtype
from ..subtyping_params import SubtypingParams
from ..qc.const import QC
from ..qc.checks import \
    is_missing_tiles, \
    is_mixed_subtype, \
    is_maybe_intermediate_subtype, \
    is_missing_too_many_target_sites, \
    is_missing_downstream_targets, \
    is_overall_coverage_low


CHECKS = [is_missing_tiles,
          is_mixed_subtype,
          is_missing_too_many_target_sites,
          is_missing_downstream_targets,
          is_maybe_intermediate_subtype,
          is_overall_coverage_low
          ] # type: List[Callable[[Subtype, DataFrame, SubtypingParams], Tuple[str, str]]]


def perform_quality_check(st: Subtype, df: DataFrame, subtyping_params: SubtypingParams) -> Tuple[str, str]:
    """Perform QC of subtyping results

    Return immediate fail if subtype result is missing or if there are no detailed subtyping results.

    Args:
        st: Subtyping results.
        df: DataFrame containing subtyping results.
        p: Subtyping/QC parameters

    Returns:
        (QC status, QC messages)
    """
    if st.subtype is None or len(st.subtype) == 0 \
            or df is None or df.shape[0] == 0:
        return QC.FAIL, QC.NO_SUBTYPE_RESULT

    overall_qc_status = QC.PASS
    messages = []
    for func in CHECKS:
        status, message = func(st, df, subtyping_params)
        # If quality check function passes, move on to the next.
        if status is None:
            continue
        messages.append('{}: {}'.format(status, message))
        if overall_qc_status == QC.FAIL:
            continue
        if status == QC.FAIL:
            overall_qc_status = QC.FAIL
            continue
        if status == QC.WARNING:
            overall_qc_status = QC.WARNING

    return overall_qc_status, ' | '.join(messages)



from typing import Tuple, Optional, List, Any, Dict

from pandas import DataFrame

from ..subtype import Subtype


def get_conflicting_tiles(st: Subtype, df: DataFrame) -> Optional[DataFrame]:
    """ Get positive and negative tiles that both are present for a subtype.

    Find positive and negative tiles for the same refposition/target site in the results `df`.

    Args:
        st: Subtype results summary
        df: Detailed subtyping results

    Returns:
        DataFrame of conflicting positive and negative tiles
    """
    if st.is_fastq_input():
        dfst = df[(df['subtype'] == str(st.subtype)) & (df['is_kmer_freq_okay'])]
    else:  # fasta files
        dfst = df[(df['subtype'] == str(st.subtype))]

    pos_tile_positions = dfst[dfst['is_pos_tile']]['refposition']
    neg_tiles = dfst[~dfst['is_pos_tile']]
    return neg_tiles[neg_tiles['refposition'].isin(pos_tile_positions)]


def get_num_pos_neg_tiles(st: Subtype, df: DataFrame) -> Tuple[int, int]:
    """Get the number of positive and negative tiles for a given subtype

    Args:
        st: Subtype results summary
        df: Detailed subtyping results

    Returns:
        (number of positive tiles, number of negative tiles)
    """
    dfst = df[(df['subtype'] == str(st.subtype))]
    return dfst[dfst['is_pos_tile']].shape[0], dfst[~dfst['is_pos_tile']].shape[0]


def get_mixed_subtype_tile_counts(dfpos: DataFrame, subtype_list: List[Any]) -> Dict[str, int]:
    st_pos_tiles = {}

    for st in subtype_list:
        bs = dfpos.subtype == [st[:x] for x in dfpos.subtype.str.len()]
        st_pos_tiles[st] = bs.sum()

    return st_pos_tiles



# -*- coding: utf-8 -*-

import re
from pkg_resources import resource_filename

from bio_hansel import program_name
from bio_hansel.subtyping_params import SubtypingParams

SCHEME_FASTAS = {'heidelberg': {'file': resource_filename(program_name, 'data/heidelberg/tiles.fasta'),
                                'version': '0.5.0',
                                'subtyping_params': SubtypingParams(low_coverage_depth_freq=20)},
                 'enteritidis': {'file': resource_filename(program_name, 'data/enteritidis/tiles.fasta'),
                                 'version': '0.7.0',
                                 'subtyping_params': SubtypingParams(low_coverage_depth_freq=50)}}

COLUMNS_TO_REMOVE = '''
pident
length
mismatch
gapopen
qstart
qend
sstart
send
evalue
bitscore
qlen
slen
sseq
is_trunc
coverage
'''.strip().split('\n')

# These are present within the subtype module.
SUBTYPE_SUMMARY_COLS = """
sample
scheme
scheme_version
subtype
all_subtypes
tiles_matching_subtype
are_subtypes_consistent
inconsistent_subtypes
n_tiles_matching_all
n_tiles_matching_all_expected
n_tiles_matching_positive
n_tiles_matching_positive_expected
n_tiles_matching_subtype
n_tiles_matching_subtype_expected
file_path
avg_tile_coverage
qc_status
qc_message
""".strip().split('\n')


SIMPLE_SUMMARY_COLS = """
sample
subtype
coverage
qc_status
qc_message
""".strip().split('\n')

REGEX_FASTQ = re.compile(r'^(.+)\.(fastq|fq|fastqsanger)(\.gz)?$')
REGEX_FASTA = re.compile(r'^.+\.(fasta|fa|fna|fas)(\.gz)?$')

JSON_EXT_TMPL = '{}.json'